@@ -1,6 +1,6 @@
 /* Multi-purpose tool for tar and cpio testsuite.
 
-   Copyright (C) 1995-1997, 2001-2018 Free Software Foundation, Inc.
+   Copyright (C) 1995-1997, 2001-2018, 2023 Free Software Foundation, Inc.
 
    François Pinard <pinard@iro.umontreal.ca>, 1995.
    Sergey Poznyakoff <gray@gnu.org>, 2004-2018.
@@ -26,7 +26,6 @@
 #include <argp.h>
 #include <argcv.h>
 #include <parse-datetime.h>
-#include <inttostr.h>
 #include <fcntl.h>
 #include <sys/stat.h>
 #include <c-ctype.h>
@@ -40,6 +39,8 @@
 #ifndef EXIT_FAILURE
 # define EXIT_FAILURE 1
 #endif
+#define EXIT_USAGE 2
+#define EXIT_UNAVAILABLE 3
 
 #if ! defined SIGCHLD && defined SIGCLD
 # define SIGCHLD SIGCLD
@@ -69,14 +70,15 @@ static off_t seek_offset = 0;
 static enum pattern pattern = DEFAULT_PATTERN;
 
 /* Next checkpoint number */
-size_t checkpoint;
+uintmax_t checkpoint;
 
 enum genfile_mode
   {
     mode_generate,
     mode_sparse,
     mode_stat,
-    mode_exec
+    mode_exec,
+    mode_set_times
   };
 
 enum genfile_mode mode = mode_generate;
@@ -105,6 +107,9 @@ int verbose;
 /* Quiet mode */
 int quiet;
 
+/* Don't dereference symlinks (for --stat) */
+int no_dereference_option;
+
 const char *argp_program_version = "genfile (" PACKAGE ") " VERSION;
 const char *argp_program_bug_address = "<" PACKAGE_BUGREPORT ">";
 static char doc[] = N_("genfile manipulates data files for GNU paxutils test suite.\n"
@@ -154,6 +159,14 @@ static struct argp_option options[] = {
    N_("Print contents of struct stat for each given file. Default FORMAT is: ")
    DEFAULT_STAT_FORMAT,
    GRP+1 },
+  {"no-dereference", 'h', NULL, 0,
+   N_("stat symbolic links instead of referenced files"),
+   GRP+1 },
+
+  {"set-times", 't', NULL, 0,
+   N_("Set access and modification times of the files to the time supplied"
+      " by --date option"),
+   GRP+1 },
 
 #undef GRP
 #define GRP 20
@@ -242,15 +255,15 @@ get_size (const char *str, int allow_zero)
       if (9 < (unsigned) digit)
 	{
 	  if (xlat_suffix (&v, p))
-	    error (EXIT_FAILURE, 0, _("Invalid size: %s"), str);
+	    error (EXIT_USAGE, 0, _("Invalid size: %s"), str);
 	  else
 	    break;
 	}
       else if (x / 10 != v)
-	error (EXIT_FAILURE, 0, _("Number out of allowed range: %s"), str);
+	error (EXIT_USAGE, 0, _("Number out of allowed range: %s"), str);
       v = x + digit;
       if (v < 0)
-	error (EXIT_FAILURE, 0, _("Negative size: %s"), str);
+	error (EXIT_USAGE, 0, _("Negative size: %s"), str);
     }
   return v;
 }
@@ -266,18 +279,21 @@ verify_file (char *file_name)
 	error (0, errno, _("stat(%s) failed"), file_name);
 
       if (st.st_size != file_length + seek_offset)
-	error (EXIT_FAILURE, 0, _("requested file length %lu, actual %lu"),
-	       (unsigned long)st.st_size, (unsigned long)file_length);
+	{
+	  intmax_t requested = st.st_size, actual = file_length;
+	  error (EXIT_FAILURE, 0, _("requested file length %jd, actual %jd"),
+		 requested, actual);
+	}
 
       if (!quiet && mode == mode_sparse && !ST_IS_SPARSE (st))
-	error (EXIT_FAILURE, 0, _("created file is not sparse"));
+	error (EXIT_UNAVAILABLE, 0, _("created file is not sparse"));
     }
 }
 
 struct action
 {
   struct action *next;
-  size_t checkpoint;
+  uintmax_t checkpoint;
   int action;
   char *name;
   off_t size;
@@ -333,7 +349,7 @@ parse_opt (int key, char *arg, struct argp_state *state)
     case 'q':
       quiet = 1;
       break;
-      
+
     case 's':
       mode = mode_sparse;
       break;
@@ -344,6 +360,14 @@ parse_opt (int key, char *arg, struct argp_state *state)
 	stat_format = arg;
       break;
 
+    case 't':
+      mode = mode_set_times;
+      break;
+
+    case 'h':
+      no_dereference_option = 1;
+      break;
+
     case 'r':
       mode = mode_exec;
       checkpoint_granularity = arg ? arg : "1";
@@ -361,7 +385,7 @@ parse_opt (int key, char *arg, struct argp_state *state)
       {
 	char *p;
 
-	checkpoint = strtoul (arg, &p, 0);
+	checkpoint = strtoumax (arg, &p, 0);
 	if (*p)
 	  argp_error (state, _("Error parsing number near `%s'"), p);
       }
@@ -534,7 +558,7 @@ make_fragment (int fd, char *offstr, char *mapstr)
 	  for (i = 0; i < block_size; i++)
 	    buffer[i] = i & 255;
 	  break;
-	  
+
 	case ZEROS_PATTERN:
 	  memset (buffer, 0, block_size);
 	  break;
@@ -542,7 +566,7 @@ make_fragment (int fd, char *offstr, char *mapstr)
 
       if (lseek (fd, displ, SEEK_CUR) == -1)
 	error (EXIT_FAILURE, errno, "lseek");
-      
+
       for (; n; n--)
 	{
 	  if (write (fd, buffer, block_size) != block_size)
@@ -565,7 +589,7 @@ generate_sparse_file (int argc, char **argv)
   int flags = O_CREAT | O_RDWR | O_BINARY;
 
   if (!file_name)
-    error (EXIT_FAILURE, 0,
+    error (EXIT_USAGE, 0,
 	   _("cannot generate sparse files on standard output, use --file option"));
   if (!seek_offset)
     flags |= O_TRUNC;
@@ -588,7 +612,7 @@ generate_sparse_file (int argc, char **argv)
 
 	      while (n > 0 && c_isspace (buf[n-1]))
 		buf[--n] = 0;
-	      
+
 	      n = strcspn (buf, " \t");
 	      buf[n++] = 0;
 	      while (buf[n] && c_isblank (buf[n]))
@@ -613,6 +637,22 @@ generate_sparse_file (int argc, char **argv)
 
 /* Status Mode */
 
+#define PRINT_INT(expr) \
+  do \
+    { \
+      if (EXPR_SIGNED (expr)) \
+	{ \
+	  intmax_t printval = expr; \
+	  printf ("%jd", printval); \
+	} \
+      else \
+	{ \
+	  uintmax_t printval = expr; \
+	  printf ("%ju", printval); \
+	} \
+    } \
+  while (false)
+
 void
 print_time (time_t t)
 {
@@ -626,9 +666,8 @@ print_stat (const char *name)
 {
   char *fmt, *p;
   struct stat st;
-  char buf[UINTMAX_STRSIZE_BOUND];
 
-  if (stat (name, &st))
+  if ((no_dereference_option ? lstat : stat) (name, &st))
     {
       error (0, errno, _("stat(%s) failed"), name);
       return;
@@ -642,53 +681,53 @@ print_stat (const char *name)
       if (strcmp (p, "name") == 0)
 	printf ("%s", name);
       else if (strcmp (p, "dev") == 0)
-	printf ("%lu", (unsigned long) st.st_dev);
+	PRINT_INT (st.st_dev);
       else if (strcmp (p, "ino") == 0)
-	printf ("%lu", (unsigned long) st.st_ino);
+	PRINT_INT (st.st_ino);
       else if (strncmp (p, "mode", 4) == 0)
 	{
-	  unsigned val = st.st_mode;
+	  uintmax_t val = st.st_mode;
 
 	  if (ispunct ((unsigned char) p[4]))
 	    {
 	      char *q;
 
-	      val &= strtoul (p + 5, &q, 8);
+	      val &= strtoumax (p + 5, &q, 8);
 	      if (*q)
 		{
 		  printf ("\n");
-		  error (EXIT_FAILURE, 0, _("incorrect mask (near `%s')"), q);
+		  error (EXIT_USAGE, 0, _("incorrect mask (near `%s')"), q);
 		}
 	    }
 	  else if (p[4])
 	    {
 	      printf ("\n");
-	      error (EXIT_FAILURE, 0, _("Unknown field `%s'"), p);
+	      error (EXIT_USAGE, 0, _("Unknown field `%s'"), p);
 	    }
-	  printf ("%0o", val);
+	  printf ("%0jo", val);
 	}
       else if (strcmp (p, "nlink") == 0)
-	printf ("%lu", (unsigned long) st.st_nlink);
+	PRINT_INT (st.st_nlink);
       else if (strcmp (p, "uid") == 0)
-	printf ("%ld", (long unsigned) st.st_uid);
+	PRINT_INT (st.st_uid);
       else if (strcmp (p, "gid") == 0)
-	printf ("%lu", (unsigned long) st.st_gid);
+	PRINT_INT (st.st_gid);
       else if (strcmp (p, "size") == 0)
-	printf ("%s", umaxtostr (st.st_size, buf));
+	PRINT_INT (st.st_size);
       else if (strcmp (p, "blksize") == 0)
-	printf ("%s", umaxtostr (st.st_blksize, buf));
+	PRINT_INT (st.st_blksize);
       else if (strcmp (p, "blocks") == 0)
-	printf ("%s", umaxtostr (st.st_blocks, buf));
+	PRINT_INT (st.st_blocks);
       else if (strcmp (p, "atime") == 0)
-	printf ("%lu", (unsigned long) st.st_atime);
+	PRINT_INT (st.st_atime);
       else if (strcmp (p, "atimeH") == 0)
 	print_time (st.st_atime);
       else if (strcmp (p, "mtime") == 0)
-	printf ("%lu", (unsigned long) st.st_mtime);
+	PRINT_INT (st.st_mtime);
       else if (strcmp (p, "mtimeH") == 0)
 	print_time (st.st_mtime);
       else if (strcmp (p, "ctime") == 0)
-	printf ("%lu", (unsigned long) st.st_ctime);
+	PRINT_INT (st.st_ctime);
       else if (strcmp (p, "ctimeH") == 0)
 	print_time (st.st_ctime);
       else if (strcmp (p, "sparse") == 0)
@@ -696,7 +735,7 @@ print_stat (const char *name)
       else
 	{
 	  printf ("\n");
-	  error (EXIT_FAILURE, 0, _("Unknown field `%s'"), p);
+	  error (EXIT_USAGE, 0, _("Unknown field `%s'"), p);
 	}
       p = strtok (NULL, ",");
       if (p)
@@ -706,6 +745,17 @@ print_stat (const char *name)
   free (fmt);
 }
 
+void
+set_times (char const *name)
+{
+  struct timespec ts[2];
+
+  ts[0] = ts[1] = touch_time;
+  if (utimensat (AT_FDCWD, name, ts, no_dereference_option ? AT_SYMLINK_NOFOLLOW : 0) != 0)
+    {
+      error (EXIT_FAILURE, errno, _("cannot set time on `%s'"), name);
+    }
+}
 
 /* Exec Mode */
 
@@ -713,7 +763,7 @@ void
 exec_checkpoint (struct action *p)
 {
   if (verbose)
-    printf ("processing checkpoint %lu\n", (unsigned long) p->checkpoint);
+    printf ("processing checkpoint %ju\n", p->checkpoint);
   switch (p->action)
     {
     case OPT_TOUCH:
@@ -721,7 +771,7 @@ exec_checkpoint (struct action *p)
 	struct timespec ts[2];
 
 	ts[0] = ts[1] = p->ts;
-	if (utimensat (AT_FDCWD, p->name, ts, 0) != 0)
+	if (utimensat (AT_FDCWD, p->name, ts, no_dereference_option ? AT_SYMLINK_NOFOLLOW : 0) != 0)
 	  {
 	    error (0, errno, _("cannot set time on `%s'"), p->name);
 	    break;
@@ -786,7 +836,7 @@ exec_checkpoint (struct action *p)
 }
 
 void
-process_checkpoint (size_t n)
+process_checkpoint (uintmax_t n)
 {
   struct action *p, *prev = NULL;
 
@@ -828,7 +878,7 @@ exec_command (int argc, char **argv)
   char **xargv;
   int i;
   char checkpoint_option[80];
-  
+
   /* Insert --checkpoint option.
      FIXME: This assumes that argv does not use traditional tar options
      (without dash).
@@ -895,7 +945,8 @@ exec_command (int argc, char **argv)
 	      && memcmp (p, CHECKPOINT_TEXT, sizeof CHECKPOINT_TEXT - 1) == 0)
 	    {
 	      char *end;
-	      size_t n = strtoul (p + sizeof CHECKPOINT_TEXT - 1, &end, 10);
+	      uintmax_t n = strtoumax (p + sizeof CHECKPOINT_TEXT - 1,
+				       &end, 10);
 	      if (!(*end && !isspace ((unsigned char) *end)))
 		{
 		  process_checkpoint (n);
@@ -952,7 +1003,7 @@ main (int argc, char **argv)
   /* Decode command options.  */
 
   if (argp_parse (&argp, argc, argv, 0, &index, NULL))
-    exit (EXIT_FAILURE);
+    exit (EXIT_USAGE);
 
   argc -= index;
   argv += index;
@@ -961,12 +1012,20 @@ main (int argc, char **argv)
     {
     case mode_stat:
       if (argc == 0)
-	error (EXIT_FAILURE, 0, _("--stat requires file names"));
+	error (EXIT_USAGE, 0, _("--stat requires file names"));
 
       while (argc--)
 	print_stat (*argv++);
       break;
 
+    case mode_set_times:
+      if (argc == 0)
+	error (EXIT_USAGE, 0, _("--set-times requires file names"));
+
+      while (argc--)
+	set_times (*argv++);
+      break;
+
     case mode_sparse:
       generate_sparse_file (argc, argv);
       verify_file (file_name);
@@ -974,7 +1033,7 @@ main (int argc, char **argv)
 
     case mode_generate:
       if (argc)
-	error (EXIT_FAILURE, 0, _("too many arguments"));
+	error (EXIT_USAGE, 0, _("too many arguments"));
       if (files_from)
 	generate_files_from_list ();
       else
